@@ -40,14 +40,10 @@
 #include <glib.h>
 #include <glib/gstdio.h>
 #include <glib/gi18n.h>
-#include <dbus/dbus.h>
-#include <dbus/dbus-glib.h>
-
-#include <nm-setting-vpn.h>
-#include <nm-utils.h>
 
 #include "nm-fortisslvpn-service.h"
 #include "nm-ppp-status.h"
+#include "nm-fortisslvpn-pppd-service-dbus.h"
 
 #if !defined(DIST_VERSION)
 # define DIST_VERSION VERSION
@@ -55,258 +51,18 @@
 
 static gboolean debug = FALSE;
 
-/********************************************************/
-/* ppp plugin <-> fortisslvpn-service object            */
-/********************************************************/
-
-/* Have to have a separate objec to handle ppp plugin requests since
- * dbus-glib doesn't allow multiple interfaces registed on one GObject.
- */
-
-#define NM_TYPE_FORTISSLVPN_PPP_SERVICE            (nm_fortisslvpn_ppp_service_get_type ())
-#define NM_FORTISSLVPN_PPP_SERVICE(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), NM_TYPE_FORTISSLVPN_PPP_SERVICE, NMFortisslvpnPppService))
-#define NM_FORTISSLVPN_PPP_SERVICE_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), NM_TYPE_FORTISSLVPN_PPP_SERVICE, NMFortisslvpnPppServiceClass))
-#define NM_IS_FORTISSLVPN_PPP_SERVICE(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), NM_TYPE_FORTISSLVPN_PPP_SERVICE))
-#define NM_IS_FORTISSLVPN_PPP_SERVICE_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), NM_TYPE_FORTISSLVPN_PPP_SERVICE))
-#define NM_FORTISSLVPN_PPP_SERVICE_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), NM_TYPE_FORTISSLVPN_PPP_SERVICE, NMFortisslvpnPppServiceClass))
-
-typedef struct {
-	GObject parent;
-} NMFortisslvpnPppService;
-
-typedef struct {
-	GObjectClass parent;
-
-	/* Signals */
-	void (*plugin_alive) (NMFortisslvpnPppService *self);
-	void (*ppp_state) (NMFortisslvpnPppService *self, guint32 state);
-	void (*ip4_config) (NMFortisslvpnPppService *self, GHashTable *config_hash);
-} NMFortisslvpnPppServiceClass;
-
-GType nm_fortisslvpn_ppp_service_get_type (void);
-
-G_DEFINE_TYPE (NMFortisslvpnPppService, nm_fortisslvpn_ppp_service, G_TYPE_OBJECT)
-
-static gboolean impl_fortisslvpn_service_set_state (NMFortisslvpnPppService *self,
-                                                    guint32 state,
-                                                    GError **err);
-
-static gboolean impl_fortisslvpn_service_set_ip4_config (NMFortisslvpnPppService *self,
-                                                         GHashTable *config,
-                                                         GError **err);
-
-#include "nm-fortisslvpn-pppd-service-glue.h"
-
-
-#define NM_FORTISSLVPN_PPP_SERVICE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NM_TYPE_FORTISSLVPN_PPP_SERVICE, NMFortisslvpnPppServicePrivate))
-
-typedef struct {
-	char *username;
-	char *password;
-} NMFortisslvpnPppServicePrivate;
-
-enum {
-	PLUGIN_ALIVE,
-	PPP_STATE,
-	IP4_CONFIG,
-
-	LAST_SIGNAL
-};
-static guint signals[LAST_SIGNAL] = { 0 };
-
-static gboolean
-_service_cache_credentials (NMFortisslvpnPppService *self,
-                            NMConnection *connection,
-                            GError **error)
-{
-	NMFortisslvpnPppServicePrivate *priv = NM_FORTISSLVPN_PPP_SERVICE_GET_PRIVATE (self);
-	NMSettingVPN *s_vpn;
-	const char *username, *password;
-
-	g_return_val_if_fail (self != NULL, FALSE);
-	g_return_val_if_fail (connection != NULL, FALSE);
-
-	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
-	if (!s_vpn) {
-		g_set_error_literal (error,
-		                     NM_VPN_PLUGIN_ERROR,
-		                     NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
-		                     _("Could not find secrets (connection invalid, no vpn setting)."));
-		return FALSE;
-	}
-
-	/* Username; try SSLVPN specific username first, then generic username */
-	username = nm_setting_vpn_get_data_item (s_vpn, NM_FORTISSLVPN_KEY_USER);
-	if (!username || !strlen (username)) {
-		username = nm_setting_vpn_get_user_name (s_vpn);
-		if (!username || !strlen (username)) {
-			g_set_error_literal (error,
-			                     NM_VPN_PLUGIN_ERROR,
-			                     NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
-			                     _("Missing VPN username."));
-			return FALSE;
-		}
-	}
-
-	password = nm_setting_vpn_get_secret (s_vpn, NM_FORTISSLVPN_KEY_PASSWORD);
-	if (!password || !strlen (password)) {
-		g_set_error_literal (error,
-		                     NM_VPN_PLUGIN_ERROR,
-		                     NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
-		                     _("Missing or invalid VPN password."));
-		return FALSE;
-	}
-
-	priv->username = g_strdup (username);
-	priv->password = g_strdup (password);
-	return TRUE;
-}
-
-static NMFortisslvpnPppService *
-nm_fortisslvpn_ppp_service_new (NMConnection *connection, GError **error)
-{
-	NMFortisslvpnPppService *self = NULL;
-	DBusGConnection *bus;
-	DBusGProxy *proxy;
-	gboolean success = FALSE;
-	guint result;
-
-	bus = dbus_g_bus_get (DBUS_BUS_SYSTEM, error);
-	if (!bus)
-		return NULL;
-	dbus_connection_set_change_sigpipe (TRUE);
-
-	proxy = dbus_g_proxy_new_for_name (bus,
-	                                   "org.freedesktop.DBus",
-	                                   "/org/freedesktop/DBus",
-	                                   "org.freedesktop.DBus");
-	if (!proxy)
-		goto out;
-	success = dbus_g_proxy_call (proxy, "RequestName", error,
-	                             G_TYPE_STRING, NM_DBUS_SERVICE_FORTISSLVPN_PPP,
-	                             G_TYPE_UINT, 0,
-	                             G_TYPE_INVALID,
-	                             G_TYPE_UINT, &result,
-	                             G_TYPE_INVALID);
-	g_object_unref (proxy);
-	if (success == FALSE)
-		goto out;
-
-	self = (NMFortisslvpnPppService *) g_object_new (NM_TYPE_FORTISSLVPN_PPP_SERVICE, NULL);
-	g_assert (self);
-
-	/* Cache the username and password so we can relay the secrets to the pppd
-	 * plugin when it asks for them.
-	 */
-	if (!_service_cache_credentials (self, connection, error)) {
-		g_clear_object (&self);
-		goto out;
-	}
-
-	dbus_g_connection_register_g_object (bus, NM_DBUS_PATH_FORTISSLVPN_PPP, G_OBJECT (self));
-
-out:
-	dbus_g_connection_unref (bus);
-	return self;
-}
-
-static void
-nm_fortisslvpn_ppp_service_init (NMFortisslvpnPppService *self)
-{
-}
-
-static void
-finalize (GObject *object)
-{
-	NMFortisslvpnPppServicePrivate *priv = NM_FORTISSLVPN_PPP_SERVICE_GET_PRIVATE (object);
-
-	/* Get rid of the cached username and password */
-	g_free (priv->username);
-	if (priv->password) {
-		memset (priv->password, 0, strlen (priv->password));
-		g_free (priv->password);
-	}
-}
-
-static void
-nm_fortisslvpn_ppp_service_class_init (NMFortisslvpnPppServiceClass *service_class)
-{
-	GObjectClass *object_class = G_OBJECT_CLASS (service_class);
-
-	g_type_class_add_private (service_class, sizeof (NMFortisslvpnPppServicePrivate));
-
-	/* virtual methods */
-	object_class->finalize = finalize;
-
-	/* Signals */
-	signals[PLUGIN_ALIVE] = 
-		g_signal_new ("plugin-alive", 
-		              G_OBJECT_CLASS_TYPE (object_class),
-		              G_SIGNAL_RUN_FIRST,
-		              G_STRUCT_OFFSET (NMFortisslvpnPppServiceClass, plugin_alive),
-		              NULL, NULL,
-		              g_cclosure_marshal_VOID__VOID,
-		              G_TYPE_NONE, 0);
-
-	signals[PPP_STATE] = 
-		g_signal_new ("ppp-state", 
-		              G_OBJECT_CLASS_TYPE (object_class),
-		              G_SIGNAL_RUN_FIRST,
-		              G_STRUCT_OFFSET (NMFortisslvpnPppServiceClass, ppp_state),
-		              NULL, NULL,
-		              g_cclosure_marshal_VOID__UINT,
-		              G_TYPE_NONE, 1, G_TYPE_UINT);
-
-	signals[IP4_CONFIG] = 
-		g_signal_new ("ip4-config", 
-		              G_OBJECT_CLASS_TYPE (object_class),
-		              G_SIGNAL_RUN_FIRST,
-		              G_STRUCT_OFFSET (NMFortisslvpnPppServiceClass, ip4_config),
-		              NULL, NULL,
-		              g_cclosure_marshal_VOID__POINTER,
-		              G_TYPE_NONE, 1, G_TYPE_POINTER);
-
-	dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (service_class),
-	                                 &dbus_glib_nm_fortisslvpn_pppd_service_object_info);
-}
-
-static gboolean
-impl_fortisslvpn_service_set_state (NMFortisslvpnPppService *self,
-                                    guint32 pppd_state,
-                                    GError **err)
-{
-	g_signal_emit (G_OBJECT (self), signals[PLUGIN_ALIVE], 0);
-	g_signal_emit (G_OBJECT (self), signals[PPP_STATE], 0, pppd_state);
-	return TRUE;
-}
-
-static gboolean
-impl_fortisslvpn_service_set_ip4_config (NMFortisslvpnPppService *self,
-                                         GHashTable *config_hash,
-                                         GError **err)
-{
-	g_message ("FORTISSLVPN service (IP Config Get) reply received.");
-	g_signal_emit (G_OBJECT (self), signals[PLUGIN_ALIVE], 0);
-
-	/* Just forward the pppd plugin config up to our superclass; no need to modify it */
-	g_signal_emit (G_OBJECT (self), signals[IP4_CONFIG], 0, config_hash);
-
-	return TRUE;
-}
-
-
 /********************************************************/
 /* The VPN plugin service                               */
 /********************************************************/
 
-G_DEFINE_TYPE (NMFortisslvpnPlugin, nm_fortisslvpn_plugin, NM_TYPE_VPN_PLUGIN)
+G_DEFINE_TYPE (NMFortisslvpnPlugin, nm_fortisslvpn_plugin, NM_TYPE_VPN_SERVICE_PLUGIN)
 
 typedef struct {
 	GPid pid;
 	guint32 ppp_timeout_handler;
-	NMFortisslvpnPppService *service;
 	NMConnection *connection;
 	char *config_file;
+	NMDBusFortisslvpnPpp *dbus_skeleton;
 } NMFortisslvpnPluginPrivate;
 
 #define NM_FORTISSLVPN_PLUGIN_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NM_TYPE_FORTISSLVPN_PLUGIN, NMFortisslvpnPluginPrivate))
@@ -465,7 +221,7 @@ validate_one_property (const char *key, const char *value, gpointer user_data)
 }
 
 static gboolean
-validate_properties (NMSettingVPN *s_vpn, GError **error)
+validate_properties (NMSettingVpn *s_vpn, GError **error)
 {
 	ValidateInfo info = { &valid_properties[0], error, FALSE };
 	int i;
@@ -506,7 +262,7 @@ validate_properties (NMSettingVPN *s_vpn, GError **error)
 }
 
 static gboolean
-validate_secrets (NMSettingVPN *s_vpn, GError **error)
+validate_secrets (NMSettingVpn *s_vpn, GError **error)
 {
 	ValidateInfo info = { &valid_secrets[0], error, FALSE };
 
@@ -550,7 +306,6 @@ cleanup_plugin (NMFortisslvpnPlugin *plugin)
 	}
 
 	g_clear_object (&priv->connection);
-	g_clear_object (&priv->service);
 	if (priv->config_file) {
 		g_unlink (priv->config_file);
 		g_clear_pointer (&priv->config_file, g_free);
@@ -564,11 +319,8 @@ openfortivpn_watch_cb (GPid pid, gint status, gpointer user_data)
 	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
 	guint error = 0;
 
-	if (WIFEXITED (status)) {
+	if (WIFEXITED (status))
 		error = WEXITSTATUS (status);
-		if (error != 0)
-			g_warning ("openfortivpn exited with error code %d", error);
-	}
 	else if (WIFSTOPPED (status))
 		g_warning ("openfortivpn stopped unexpectedly with signal %d", WSTOPSIG (status));
 	else if (WIFSIGNALED (status))
@@ -582,11 +334,10 @@ openfortivpn_watch_cb (GPid pid, gint status, gpointer user_data)
 
 	/* TODO: Better error indication from openfortivpn. */
 	if (error)
-		nm_vpn_plugin_failure (NM_VPN_PLUGIN (plugin),
-		                       NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
-
-	nm_vpn_plugin_set_state (NM_VPN_PLUGIN (plugin), NM_VPN_SERVICE_STATE_STOPPED);
-	cleanup_plugin (plugin);
+		nm_vpn_service_plugin_failure (NM_VPN_SERVICE_PLUGIN (plugin),
+		                               NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
+	else
+		nm_vpn_service_plugin_disconnect (NM_VPN_SERVICE_PLUGIN (plugin), NULL);
 }
 
 static const char *
@@ -617,13 +368,13 @@ pppd_timed_out (gpointer user_data)
 	NMFortisslvpnPlugin *plugin = NM_FORTISSLVPN_PLUGIN (user_data);
 
 	g_warning ("Looks like pppd didn't initialize our dbus module");
-	nm_vpn_plugin_failure (NM_VPN_PLUGIN (plugin), NM_VPN_CONNECTION_STATE_REASON_SERVICE_START_TIMEOUT);
+	nm_vpn_service_plugin_failure (NM_VPN_SERVICE_PLUGIN (plugin), NM_VPN_CONNECTION_STATE_REASON_SERVICE_START_TIMEOUT);
 
 	return FALSE;
 }
 
 static gboolean
-run_openfortivpn (NMFortisslvpnPlugin *plugin, NMSettingVPN *s_vpn, GError **error)
+run_openfortivpn (NMFortisslvpnPlugin *plugin, NMSettingVpn *s_vpn, GError **error)
 {
 	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
 	GPid pid;
@@ -713,79 +464,74 @@ remove_timeout_handler (NMFortisslvpnPlugin *plugin)
 	}
 }
 
-static void
-service_plugin_alive_cb (NMFortisslvpnPppService *service,
-                         NMFortisslvpnPlugin *plugin)
+static gboolean
+handle_set_state (NMDBusFortisslvpnPpp *object,
+                  GDBusMethodInvocation *invocation,
+                  guint arg_state,
+                  gpointer user_data)
 {
-	remove_timeout_handler (plugin);
-}
+	remove_timeout_handler (NM_FORTISSLVPN_PLUGIN (user_data));
+	if (arg_state == NM_PPP_STATUS_DEAD || arg_state == NM_PPP_STATUS_DISCONNECT)
+		nm_vpn_service_plugin_disconnect (NM_VPN_SERVICE_PLUGIN (user_data), NULL);
 
-static void
-service_ppp_state_cb (NMFortisslvpnPppService *service,
-                      guint32 ppp_state,
-                      NMFortisslvpnPlugin *plugin)
-{
-	NMVPNServiceState plugin_state = nm_vpn_plugin_get_state (NM_VPN_PLUGIN (plugin));
-
-	switch (ppp_state) {
-	case NM_PPP_STATUS_DEAD:
-	case NM_PPP_STATUS_DISCONNECT:
-		if (plugin_state == NM_VPN_SERVICE_STATE_STARTED)
-			nm_vpn_plugin_disconnect (NM_VPN_PLUGIN (plugin), NULL);
-		else if (plugin_state == NM_VPN_SERVICE_STATE_STARTING)
-			nm_vpn_plugin_failure (NM_VPN_PLUGIN (plugin), NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
-		break;
-	default:
-		break;
-	}
-}
-
-static void
-nm_gvalue_destroy (gpointer data)
-{
-	g_value_unset ((GValue *) data);
-	g_slice_free (GValue, data);
-}
-
-static GValue *
-nm_gvalue_dup (const GValue *value)
-{
-	GValue *value_dup;
-
-	value_dup = g_slice_new0 (GValue);
-	g_value_init (value_dup, G_VALUE_TYPE (value));
-	g_value_copy (value, value_dup);
-
-	return value_dup;
-}
-
-static void
-copy_hash (gpointer key, gpointer value, gpointer user_data)
-{
-	g_hash_table_insert ((GHashTable *) user_data, g_strdup (key), nm_gvalue_dup ((GValue *) value));
-}
-
-static void
-service_ip4_config_cb (NMFortisslvpnPppService *service,
-                       GHashTable *config_hash,
-                       NMVPNPlugin *plugin)
-{
-	GHashTable *hash;
-
-	hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, nm_gvalue_destroy);
-	g_hash_table_foreach (config_hash, copy_hash, hash);
-	nm_vpn_plugin_set_ip4_config (plugin, hash);
-	g_hash_table_destroy (hash);
+	nmdbus_fortisslvpn_ppp_complete_set_state (object, invocation);
+	return TRUE;
 }
 
 static gboolean
-real_connect (NMVPNPlugin *plugin, NMConnection *connection, GError **error)
+handle_set_ip4_config (NMDBusFortisslvpnPpp *object,
+                       GDBusMethodInvocation *invocation,
+                       GVariant *arg_config,
+                       gpointer user_data)
+{
+	g_message ("FORTISSLVPN service (IP Config Get) reply received.");
+
+	remove_timeout_handler (NM_FORTISSLVPN_PLUGIN (user_data));
+	nm_vpn_service_plugin_set_ip4_config (NM_VPN_SERVICE_PLUGIN (user_data), arg_config);
+
+	nmdbus_fortisslvpn_ppp_complete_set_ip4_config (object, invocation);
+	return TRUE;
+}
+
+static gboolean
+get_credentials (NMSettingVpn *s_vpn,
+                 const char **username,
+                 const char **password,
+                 GError **error)
+{
+	/* Username; try SSLVPN specific username first, then generic username */
+	*username = nm_setting_vpn_get_data_item (s_vpn, NM_FORTISSLVPN_KEY_USER);
+	if (!*username || !strlen (*username)) {
+		*username = nm_setting_vpn_get_user_name (s_vpn);
+		if (!*username || !strlen (*username)) {
+			g_set_error_literal (error,
+			                     NM_VPN_PLUGIN_ERROR,
+			                     NM_VPN_PLUGIN_ERROR_INVALID_CONNECTION,
+			                     _("Missing VPN username."));
+			return FALSE;
+		}
+	}
+
+	*password = nm_setting_vpn_get_secret (s_vpn, NM_FORTISSLVPN_KEY_PASSWORD);
+	if (!*password || !strlen (*password)) {
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_INVALID_CONNECTION,
+		                     _("Missing or invalid VPN password."));
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static gboolean
+real_connect (NMVpnServicePlugin *plugin, NMConnection *connection, GError **error)
 {
 	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
-	NMFortisslvpnPppServicePrivate *service_priv;
-	NMSettingVPN *s_vpn;
+	NMSettingVpn *s_vpn;
 	mode_t old_umask;
 	gchar *config;
+	const char *username, *password;
 
 	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
@@ -798,32 +544,21 @@ real_connect (NMVPNPlugin *plugin, NMConnection *connection, GError **error)
 	if (!validate_secrets (s_vpn, error))
 		return FALSE;
 
-	/* Start our pppd plugin helper service */
-	g_clear_object (&priv->service);
-	g_clear_object (&priv->connection);
-
-	/* Start our helper D-Bus service that the pppd plugin sends state changes to */
-	priv->service = nm_fortisslvpn_ppp_service_new (connection, error);
-	if (!priv->service)
+	if (!get_credentials (s_vpn, &username, &password, error))
 		return FALSE;
 
+	g_clear_object (&priv->connection);
 	priv->connection = g_object_ref (connection);
 
-	g_signal_connect (G_OBJECT (priv->service), "plugin-alive", G_CALLBACK (service_plugin_alive_cb), plugin);
-	g_signal_connect (G_OBJECT (priv->service), "ppp-state", G_CALLBACK (service_ppp_state_cb), plugin);
-	g_signal_connect (G_OBJECT (priv->service), "ip4-config", G_CALLBACK (service_ip4_config_cb), plugin);
-
 	if (debug)
 		nm_connection_dump (connection);
 
 	/* Create the configuration file so that we don't expose
 	 * secrets on the command line */
-	service_priv = NM_FORTISSLVPN_PPP_SERVICE_GET_PRIVATE (priv->service);
 	priv->config_file = g_strdup_printf (NM_FORTISSLVPN_STATEDIR "/%s.config",
 	                                     nm_connection_get_uuid (connection));
 	config = g_strdup_printf ("username = %s\npassword = %s\n",
-	                          service_priv->username,
-	                          service_priv->password);
+	                          username, password);
 	old_umask = umask (0077);
 	if (!g_file_set_contents (priv->config_file, config, -1, error)) {
 		g_clear_pointer (&priv->config_file, g_free);
@@ -839,15 +574,15 @@ real_connect (NMVPNPlugin *plugin, NMConnection *connection, GError **error)
 }
 
 static gboolean
-real_need_secrets (NMVPNPlugin *plugin,
+real_need_secrets (NMVpnServicePlugin *plugin,
                    NMConnection *connection,
-                   char **setting_name,
+                   const char **setting_name,
                    GError **error)
 {
 	NMSetting *s_vpn;
 	NMSettingSecretFlags flags = NM_SETTING_SECRET_FLAG_NONE;
 
-	g_return_val_if_fail (NM_IS_VPN_PLUGIN (plugin), FALSE);
+	g_return_val_if_fail (NM_IS_VPN_SERVICE_PLUGIN (plugin), FALSE);
 	g_return_val_if_fail (NM_IS_CONNECTION (connection), FALSE);
 
 	s_vpn = nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
@@ -868,14 +603,14 @@ real_need_secrets (NMVPNPlugin *plugin,
 }
 
 static gboolean
-real_disconnect (NMVPNPlugin *plugin, GError **err)
+real_disconnect (NMVpnServicePlugin *plugin, GError **err)
 {
 	cleanup_plugin (NM_FORTISSLVPN_PLUGIN (plugin));
 	return TRUE;
 }
 
 static void
-state_changed_cb (GObject *object, NMVPNServiceState state, gpointer user_data)
+state_changed_cb (GObject *object, NMVpnServiceState state, gpointer user_data)
 {
 	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (object);
 
@@ -890,7 +625,6 @@ state_changed_cb (GObject *object, NMVPNServiceState state, gpointer user_data)
 	case NM_VPN_SERVICE_STATE_STOPPED:
 		remove_timeout_handler (NM_FORTISSLVPN_PLUGIN (object));
 		g_clear_object (&priv->connection);
-		g_clear_object (&priv->service);
 		break;
 	default:
 		break;
@@ -900,7 +634,21 @@ state_changed_cb (GObject *object, NMVPNServiceState state, gpointer user_data)
 static void
 dispose (GObject *object)
 {
+	NMFortisslvpnPluginPrivate *priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (object);
+	GDBusInterfaceSkeleton *skeleton = NULL;
+
+	if (priv->dbus_skeleton)
+		skeleton = G_DBUS_INTERFACE_SKELETON (priv->dbus_skeleton);
+
 	cleanup_plugin (NM_FORTISSLVPN_PLUGIN (object));
+
+	if (skeleton) {
+		if (g_dbus_interface_skeleton_get_object_path (skeleton))
+			g_dbus_interface_skeleton_unexport (skeleton);
+		g_signal_handlers_disconnect_by_func (skeleton, handle_set_state, object);
+		g_signal_handlers_disconnect_by_func (skeleton, handle_set_ip4_config, object);
+	}
+
 	G_OBJECT_CLASS (nm_fortisslvpn_plugin_parent_class)->dispose (object);
 }
 
@@ -913,7 +661,7 @@ static void
 nm_fortisslvpn_plugin_class_init (NMFortisslvpnPluginClass *fortisslvpn_class)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (fortisslvpn_class);
-	NMVPNPluginClass *parent_class = NM_VPN_PLUGIN_CLASS (fortisslvpn_class);
+	NMVpnServicePluginClass *parent_class = NM_VPN_SERVICE_PLUGIN_CLASS (fortisslvpn_class);
 
 	g_type_class_add_private (object_class, sizeof (NMFortisslvpnPluginPrivate));
 
@@ -925,16 +673,45 @@ nm_fortisslvpn_plugin_class_init (NMFortisslvpnPluginClass *fortisslvpn_class)
 }
 
 NMFortisslvpnPlugin *
-nm_fortisslvpn_plugin_new (void)
+nm_fortisslvpn_plugin_new (const char *bus_name)
 {
 	NMFortisslvpnPlugin *plugin;
+	NMFortisslvpnPluginPrivate *priv;
+	GDBusConnection *connection;
+	GError *error = NULL;
 
-	plugin = g_object_new (NM_TYPE_FORTISSLVPN_PLUGIN,
-	                       NM_VPN_PLUGIN_DBUS_SERVICE_NAME,
-	                       NM_DBUS_SERVICE_FORTISSLVPN,
-	                       NULL);
-	if (plugin)
+	plugin = (NMFortisslvpnPlugin *) g_initable_new (NM_TYPE_FORTISSLVPN_PLUGIN, NULL, &error,
+	                                                 NM_VPN_SERVICE_PLUGIN_DBUS_SERVICE_NAME, bus_name,
+	                                                 NULL);
+	if (plugin) {
 		g_signal_connect (G_OBJECT (plugin), "state-changed", G_CALLBACK (state_changed_cb), NULL);
+	} else {
+		g_warning ("Failed to initialize a plugin instance: %s", error->message);
+		g_error_free (error);
+	}
+
+	connection = nm_vpn_service_plugin_get_connection (NM_VPN_SERVICE_PLUGIN (plugin)),
+	priv = NM_FORTISSLVPN_PLUGIN_GET_PRIVATE (plugin);
+	priv->dbus_skeleton = nmdbus_fortisslvpn_ppp_skeleton_new ();
+	if (!g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (priv->dbus_skeleton),
+	                                       connection,
+	                                       NM_DBUS_PATH_FORTISSLVPN_PPP,
+	                                       &error)) {
+		g_warning ("Failed to export helper interface: %s", error->message);
+		g_error_free (error);
+		g_clear_object (&plugin);
+		goto out;
+	}
+
+	g_dbus_connection_register_object (connection, NM_VPN_DBUS_PLUGIN_PATH,
+	                                   nmdbus_fortisslvpn_ppp_interface_info (),
+	                                   NULL, NULL, NULL, NULL);
+
+	g_signal_connect (priv->dbus_skeleton, "handle-set-state", G_CALLBACK (handle_set_state), plugin);
+	g_signal_connect (priv->dbus_skeleton, "handle-set-ip4-config", G_CALLBACK (handle_set_ip4_config), plugin);
+
+out:
+	g_clear_object (&connection);
 	return plugin;
 }
 
@@ -951,10 +728,12 @@ main (int argc, char *argv[])
 	GMainLoop *main_loop;
 	gboolean persist = FALSE;
 	GOptionContext *opt_ctx = NULL;
+	gchar *bus_name = NM_DBUS_SERVICE_FORTISSLVPN;
 
 	GOptionEntry options[] = {
 		{ "persist", 0, 0, G_OPTION_ARG_NONE, &persist, N_("Don't quit when VPN connection terminates"), NULL },
 		{ "debug", 0, 0, G_OPTION_ARG_NONE, &debug, N_("Enable verbose debug logging (may expose passwords)"), NULL },
+		{ "bus-name", 0, 0, G_OPTION_ARG_STRING, &bus_name, N_("D-Bus name to use for this instance"), NULL },
 		{NULL}
 	};
 
@@ -988,7 +767,10 @@ main (int argc, char *argv[])
 	if (debug)
 		g_message ("nm-fortisslvpn-service (version " DIST_VERSION ") starting...");
 
-	plugin = nm_fortisslvpn_plugin_new ();
+	if (bus_name)
+		setenv ("NM_DBUS_SERVICE_FORTISSLVPN", bus_name, 0);
+
+	plugin = nm_fortisslvpn_plugin_new (bus_name);
 	if (!plugin)
 		exit (EXIT_FAILURE);
 
